==> udmsearch-3.1.3.tar.gz: udmsearch-3.1.3/src/ftp.c <==
#include "udm_config.h"

#ifdef USE_FTP
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "udm_common.h"
#include "udm_socket.h"
#include "udm_host.h"   
#include "udm_utils.h"   
#include "udm_ftp.h"   
#include "udm_log.h"   
#include "udm_proto.h"   
#include "udm_xmalloc.h"   
#include "udm_conf.h"   

int ftp_get_reply(UDM_CONN *connp){
	if (!connp->buf)
		return -1;
	
        return (atoi(connp->buf)/100);
}

int ftp_read_line(UDM_CONN *connp){

	if (socket_select(connp, NET_READ_TIMEOUT, 'r')){
		UdmLog(connp->indexer->handle, UDM_LOG_DEBUG, "ftp://%s (ftp_read_line-timeout-err): ", connp->hostname);
		return -1;
	}
	
		
	do {
		if (socket_read_line(connp) < 0)
			return -1;
		if (((connp->buf[0] =='1')||(connp->buf[0] =='2')||
		    (connp->buf[0] =='3')||(connp->buf[0] =='4')||
		    (connp->buf[0] =='5')) && (connp->buf[3] == ' '))
			break;
	}while( 1 );
	return 0;
}

int ftp_connect(UDM_CONN *connp, char *hostname, int port, char *user, 
		char *passwd, int timeout){
	int len;
	
	if (!connp)
		return -1;

	if (connp->connected == NET_CONNECTED)
		ftp_close(connp);	
	connp->connected = NET_NOTCONNECTED;

	if (!port)
		connp->port = 21;
	else    
		connp->port = port;

	connp->timeout = timeout;
	
	if (!hostname)
		return -1;
	len = strlen(hostname);
	connp->hostname = UdmXrealloc(connp->hostname, len+1);
	udm_snprintf(connp->hostname, len+1, "%s", hostname);
	
	if (ftp_open_control_port(connp))
		return -1;
	if (ftp_login(connp, user, passwd))
		return -1;
	ftp_set_binary(connp);
	connp->connected = NET_CONNECTED;
	return 0;
}

int ftp_open_control_port(UDM_CONN *connp){
	int code;

        if (host_lookup(connp))
		return -1;
        if (socket_open(connp))
		return -1;
	if (socket_connect(connp))
		return -1;
	/* Read server response */
	ftp_read_line(connp);
	code = ftp_get_reply(connp);
	if (code != 2)
		return -1;
	return 0;
}

int ftp_open_data_port( UDM_CONN *c, UDM_CONN *d){
	char buf[64];	
	unsigned char *a, *p;
	
	int code;
	
	if (!d)
		return -1;
	if (socket_getname(c, &d->sin) == -1)
		return -1;

	if (socket_open(d))
		return -1;
	if (socket_listen(d)){
		return -1;
	}
	if (socket_getname(d, &d->sin) == -1){
		return -1;
	}

        a = (unsigned char *)&d->sin.sin_addr;
        p = (unsigned char *)&d->sin.sin_port;

	udm_snprintf(buf, 64, "PORT %d,%d,%d,%d,%d,%d",
                a[0], a[1], a[2], a[3], p[0], p[1]);
	code = ftp_send_cmd(c, buf);
	if (strncasecmp(c->buf, "200 PORT command successful.", 28)){
		return -1;
	}
	return 0;
}


int ftp_send_cmd(UDM_CONN *connp, char *cmd) {
        char *buf;
	int len;

    	connp->err = 0;
	len = strlen(cmd)+2;    
        buf = UdmXmalloc(len+1);
	udm_snprintf(buf, len+1, "%s\r\n", cmd);
	socket_buf_clear(connp);
	if (socket_write(connp, buf)){
		UDM_FREE( buf);
		return -1;
	}
	UdmLog(connp->indexer->handle, UDM_LOG_DEBUG, "ftp://%s (cmd) : %s", connp->hostname, buf);
	UDM_FREE(buf);
	if (ftp_read_line(connp))
		return -1;
	UdmLog(connp->indexer->handle, UDM_LOG_DEBUG, "ftp://%s (reply): %s", connp->hostname, connp->buf);
    	return(ftp_get_reply(connp));
}

int ftp_expect_bytes(char *buf){
	char *ch, *ch1;
	int bytes;
	
        ch = strstr(buf, " bytes");
        ch1 = strrchr(buf, '(');
        if (ch==0 || ch1 ==0)
                return -1;
        bytes = atol(ch1+1);

	return bytes;
}

int ftp_send_data_cmd(UDM_CONN *c, UDM_CONN *d, char *cmd){
	int code, bytes=0;
	
	if (!d)
		return -1;

	d->timeout = c->timeout;
	d->hostname = c->hostname;
	c->err = 0;
	
	if (ftp_open_data_port(c, d)){
		socket_close(d);
		return -1;
	}

	code = ftp_send_cmd(c, cmd);
	if (code == -1){
                socket_close(d);
                return -1;
	}else if ( code >3 ){
		c->err = code;
		socket_close(d);
		return -1;
	}
	bytes = ftp_expect_bytes(c->buf);	
	if (socket_accept(d)){
		socket_close(d);
		return -1;
	}

	if (socket_read(d, UdmMaxDocSize()) < 0){
		UdmLog(c->indexer->handle, UDM_LOG_DEBUG, "ftp://%s (socket_read-err):", c->hostname);
		socket_close(d);
		ftp_read_line(c);
		return -1;
	}
        socket_close(d);
	/* Check if file too lage ABORT
	if (d->err == UDM_NET_FILE_TL){
		if (ftp_abort(c)){
			socket_buf_clear(d);
			return -1;
		}
	} 
*/	/* 226 Transfer complete. */
	if (ftp_read_line(c)){
		/* Fixme: What to do if not recive "226 Transfer complete."? */
		UdmLog(c->indexer->handle, UDM_LOG_DEBUG, "ftp://%s (data-end-err): %d", d->hostname, d->buf_len);
		ftp_close(c);
		if (bytes == d->buf_len )
			return 0;
		return -1;
	}
	code = ftp_get_reply(c);
	if (code == -1){
                return -1;
	}else if ( code >3 ){
		c->err = code;
		return -1;
	}
	return 0;		
}

int ftp_login(UDM_CONN *connp, char *user, char *passwd){
	char *buf;
	char user_tmp[32], passwd_tmp[64];
	int len, code;
	
	if (!user)
		udm_snprintf(user_tmp, 32, "anonymous");
	else
		udm_snprintf(user_tmp, 32, "%s", user);
	    
	if (!passwd)
		udm_snprintf(passwd_tmp, 64, "%s-%s@search.mnogo.ru", PACKAGE, VERSION);
	else
		udm_snprintf(passwd_tmp, 32, "%s", passwd);
	
	len = strlen(user_tmp)+strlen("USER ");
	buf = (char *) UdmXmalloc(len+1);
	udm_snprintf(buf, len+1, "USER %s", user_tmp);
	code = ftp_send_cmd(connp, buf);
	UDM_FREE(buf);
	if (code == -1)
		return -1;
	else if (code == 2) /* Don't need password */
		return 0;
		
	len = strlen(passwd_tmp)+strlen("PASS ");
	buf = (char *) UdmXmalloc(len+1);
	udm_snprintf(buf, len+1, "PASS %s", passwd_tmp);
	code = ftp_send_cmd( connp, buf);
	UDM_FREE(buf);
	if (code > 3)
		return -1;
	return 0;
}

int ftp_parse_list(UDM_CONN *connp, char *path){
	char *line, *buf_in, *ch, *buf_out, *tok, *fname;
        int len, len_h, len_f,len_p, cur_len, buf_len, i;
	
	if (!connp->buf || !connp->buf_len)
		return 0;
        buf_in = connp->buf;
        len_h = strlen(connp->hostname)+22;
        len_p = strlen(path);
        cur_len = 0;
        buf_len = NET_BUF_SIZE;
        buf_out = UdmXmalloc(buf_len);
	line=UdmGetToken(buf_in,"\r\n",&tok);
        do{
    		if (!(fname = strtok(line, " ")))
			continue;
		/* drwxrwxrwx x user group size month date time file_name */
		for(i=0; i<7; i++)
            		if (!(fname = strtok(NULL, " ")))
				break;
		if (!(fname = strtok(NULL, "")))
			continue;
		len = 0 ;
		len_f = len_h + len_p + strlen(fname);
	        if ((cur_len+len_f) >= buf_len){
			buf_len += NET_BUF_SIZE;
			buf_out = UdmXrealloc(buf_out, buf_len);
		}
					
		switch (line[0]){
			case 'd':
				if (!fname || !strcmp(fname, ".") || !strcmp(fname, ".."))
				        break;
				len = len_f;
    				udm_snprintf(buf_out+cur_len, len+1, "<a href=\"ftp://%s%s%s/\"></a>",
					    connp->hostname, path, fname);
				break;
    	    	        case 'l':
				ch = strstr (fname, " -> ");
				if (!ch)
				    break;
				ch +=4;
				len = len_h+len_p+strlen(ch);
    				udm_snprintf(buf_out+cur_len, len+1, "<a  href=\"ftp://%s%s%s\"></a>", 
					    connp->hostname, path, ch);
				break;
    	    	        case '-':
				len =  len_f; 
		    	        udm_snprintf(buf_out+cur_len, len+1, "<a  href=\"ftp://%s%s%s\"></a>", 
					    connp->hostname, path, fname);

				break;
		}
		cur_len += len;
		
	}while( (line=UdmGetToken(NULL, "\r\n", &tok)));

	if (cur_len+1 > connp->buf_len_total){
		connp->buf_len_total = cur_len;  
		connp->buf = UdmXrealloc(connp->buf, connp->buf_len_total+1);
	}
	bzero(connp->buf, connp->buf_len_total+1);
	memcpy(connp->buf, buf_out, cur_len);
	UDM_FREE(buf_out);
	return 0;
}

int ftp_list(UDM_CONN *c, UDM_CONN *d, char *path, char *filename){
	char *cmd;
	int len;
	
	if (!filename){
		cmd = UdmXmalloc(5);
		sprintf(cmd, "LIST");
	}else{
    		len = strlen(filename)+5;
		cmd = UdmXmalloc(len+1);
    		udm_snprintf(cmd, len+1, "LIST %s", filename);
	}

	if (ftp_send_data_cmd(c, d, cmd)== -1){
		UDM_FREE(cmd);
		UdmLog(c->indexer->handle, UDM_LOG_DEBUG, "(ftp_list-err)->%s", d->buf);
		return -1;
	}
	UDM_FREE(cmd);
	ftp_parse_list(d, path);
	return 0;
}

int ftp_get(UDM_CONN *c, UDM_CONN *d, char *path){
	char *cmd;
	int len;
	
        if (!path)
		return -1;

        len = strlen(path)+5;
	cmd = UdmXmalloc(len+1);
        udm_snprintf(cmd, len+1, "RETR %s", path);

	if (ftp_send_data_cmd(c, d, cmd) == -1){
		UDM_FREE(cmd);
		return -1;
	}
	UDM_FREE(cmd);
	return 0;
}

int ftp_mdtm(UDM_CONN *c, char *path){
	char *cmd;
	int len, code;
	
	if (!path)
		return -1;
		
    	len = strlen(path)+5;
	cmd = UdmXmalloc(len+1);
	udm_snprintf(cmd, len+1, "MDTM %s", path);

	code = ftp_send_cmd(c, cmd);
	UDM_FREE(cmd);
	if (code == -1){
                return -1;
	}else if ( code >3 ){
		c->err = code;
		return -1;
	}
	return (UdmFTPDate2Time_t(c->buf));
}

int ftp_set_binary(UDM_CONN *c){
	char *cmd;
	int code;
	
	cmd = UdmXmalloc(7);
	sprintf(cmd, "TYPE I");

	code = ftp_send_cmd(c, cmd);
	UDM_FREE(cmd);
	if (code == -1){
                return -1;
	}else if ( code >3 ){
		c->err = code;
		return -1;
	}
	return 0;
}

int ftp_cwd(UDM_CONN *c, char *path){
	char *cmd;
	int len, code;
	
	if (!path)
		return -1;
		
	len = strlen(path)+4;
	cmd = UdmXmalloc(len+1);
	udm_snprintf(cmd, len+1, "CWD %s", path);

	code = ftp_send_cmd(c, cmd);
	UDM_FREE(cmd);
	if (code == -1){
                return -1;
	}else if ( code >3 ){
		c->err = code;
		return -1;
	}
	return 0;
}

int ftp_close(UDM_CONN *connp){
	int code;

	if (connp->connected == NET_CONNECTED){
    		code = ftp_send_cmd(connp, "QUIT");
	}
	connp->connected = NET_NOTCONNECTED;
	socket_close(connp);
	if (connp->connp)
		socket_close(connp->connp);
	return 0;
}

int ftp_abort(UDM_CONN *connp){
	int code;

	socket_buf_clear(connp->connp);
	if (send(connp->conn_fd, "\xFF\xF4\xFF", 3, MSG_OOB)==-1)
		return -1;
	
	if (socket_write(connp, "\xF2"))
		return -1;
	
	code = ftp_send_cmd(connp, "ABOR");
	socket_buf_clear(connp->connp);

	if (code !=4)
		return -1;
	return 0;
}
#endif

==> udmsearch-3.1.5.tar.gz: udmsearch-3.1.5/src/ftp.c <==
#include "udm_config.h"

#ifdef USE_FTP
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "udm_common.h"
#include "udm_socket.h"
#include "udm_host.h"   
#include "udm_utils.h"   
#include "udm_ftp.h"   
#include "udm_log.h"   
#include "udm_proto.h"   
#include "udm_xmalloc.h"   
#include "udm_conf.h"   

int ftp_get_reply(UDM_CONN *connp){
	if (!connp->buf)
		return -1;
	
        return (atoi(connp->buf)/100);
}

int ftp_read_line(UDM_CONN *connp){

	if (socket_select(connp, NET_READ_TIMEOUT, 'r')){
		UdmLog(connp->indexer, UDM_LOG_DEBUG, "ftp://%s (ftp_read_line-timeout-err): ", connp->hostname);
		return -1;
	}
	
		
	do {
		if (socket_read_line(connp) < 0)
			return -1;
		if (((connp->buf[0] =='1')||(connp->buf[0] =='2')||
		    (connp->buf[0] =='3')||(connp->buf[0] =='4')||
		    (connp->buf[0] =='5')) && (connp->buf[3] == ' '))
			break;
	}while( 1 );
	return 0;
}

int ftp_connect(UDM_ENV * Conf,UDM_CONN *connp, char *hostname, int port, char *user, 
		char *passwd, int timeout){
	int len;
	
	if (!connp)
		return -1;

	if (connp->connected == NET_CONNECTED)
		ftp_close(connp);	
	connp->connected = NET_NOTCONNECTED;

	if (!port)
		connp->port = 21;
	else    
		connp->port = port;

	connp->timeout = timeout;
	
	if (!hostname)
		return -1;
	len = strlen(hostname);
	connp->hostname = UdmXrealloc(connp->hostname, len+1);
	udm_snprintf(connp->hostname, len+1, "%s", hostname);
	
	if (ftp_open_control_port(Conf,connp))
		return -1;
	if (ftp_login(connp, user, passwd))
		return -1;
	ftp_set_binary(connp);
	connp->connected = NET_CONNECTED;
	return 0;
}

int ftp_open_control_port(UDM_ENV * Conf,UDM_CONN *connp){
	int code;

        if (UdmHostLookup(Conf,connp))
		return -1;
        if (socket_open(connp))
		return -1;
	if (socket_connect(connp))
		return -1;
	/* Read server response */
	ftp_read_line(connp);
	code = ftp_get_reply(connp);
	if (code != 2)
		return -1;
	return 0;
}

int ftp_open_data_port( UDM_CONN *c, UDM_CONN *d){
	char buf[64];	
	unsigned char *a, *p;
	
	int code;
	
	if (!d)
		return -1;
	if (socket_getname(c, &d->sin) == -1)
		return -1;

	if (socket_open(d))
		return -1;
	if (socket_listen(d)){
		return -1;
	}
	if (socket_getname(d, &d->sin) == -1){
		return -1;
	}

        a = (unsigned char *)&d->sin.sin_addr;
        p = (unsigned char *)&d->sin.sin_port;

	udm_snprintf(buf, 64, "PORT %d,%d,%d,%d,%d,%d",
                a[0], a[1], a[2], a[3], p[0], p[1]);
	code = ftp_send_cmd(c, buf);
	if (strncasecmp(c->buf, "200 PORT command successful.", 28)){
		return -1;
	}
	return 0;
}


int ftp_send_cmd(UDM_CONN *connp, char *cmd) {
        char *buf;
	int len;

    	connp->err = 0;
	len = strlen(cmd)+2;    
        buf = UdmXmalloc(len+1);
	udm_snprintf(buf, len+1, "%s\r\n", cmd);
	socket_buf_clear(connp);
	if (socket_write(connp, buf)){
		UDM_FREE( buf);
		return -1;
	}
	UdmLog(connp->indexer, UDM_LOG_DEBUG, "ftp://%s (cmd) : %s", connp->hostname, buf);
	UDM_FREE(buf);
	if (ftp_read_line(connp))
		return -1;
	UdmLog(connp->indexer, UDM_LOG_DEBUG, "ftp://%s (reply): %s", connp->hostname, connp->buf);
    	return(ftp_get_reply(connp));
}

int ftp_expect_bytes(char *buf){
	char *ch, *ch1;
	int bytes;
	
        ch = strstr(buf, " bytes");
        ch1 = strrchr(buf, '(');
        if (ch==0 || ch1 ==0)
                return -1;
        bytes = atol(ch1+1);

	return bytes;
}

int ftp_send_data_cmd(UDM_CONN *c, UDM_CONN *d, char *cmd, int max_doc_size){
	int code, bytes=0;
	
	if (!d)
		return -1;

	d->timeout = c->timeout;
	d->hostname = c->hostname;
	c->err = 0;
	
	if (ftp_open_data_port(c, d)){
		socket_close(d);
		return -1;
	}

	code = ftp_send_cmd(c, cmd);
	if (code == -1){
                socket_close(d);
                return -1;
	}else if ( code >3 ){
		c->err = code;
		socket_close(d);
		return -1;
	}
	bytes = ftp_expect_bytes(c->buf);	
	if (socket_accept(d)){
		socket_close(d);
		return -1;
	}

	if (socket_read(d, max_doc_size) < 0){
		UdmLog(c->indexer, UDM_LOG_DEBUG, "ftp://%s (socket_read-err):", c->hostname);
		socket_close(d);
		ftp_read_line(c);
		return -1;
	}
        socket_close(d);
	/* Check if file too lage ABORT
	if (d->err == UDM_NET_FILE_TL){
		if (ftp_abort(c)){
			socket_buf_clear(d);
			return -1;
		}
	} 
*/	/* 226 Transfer complete. */
	if (ftp_read_line(c)){
		/* Fixme: What to do if not receive "226 Transfer complete."? */
		UdmLog(c->indexer, UDM_LOG_DEBUG, "ftp://%s (data-end-err): %d", d->hostname, d->buf_len);
		ftp_close(c);
		if (bytes == d->buf_len )
			return 0;
		return -1;
	}
	code = ftp_get_reply(c);
	if (code == -1){
                return -1;
	}else if ( code >3 ){
		c->err = code;
		return -1;
	}
	return 0;		
}

int ftp_login(UDM_CONN *connp, char *user, char *passwd){
	char *buf;
	char user_tmp[32], passwd_tmp[64];
	int len, code;
	
	if (!user)
		udm_snprintf(user_tmp, 32, "anonymous");
	else
		udm_snprintf(user_tmp, 32, "%s", user);
	    
	if (!passwd)
		udm_snprintf(passwd_tmp, 64, "%s-%s@search.mnogo.ru", PACKAGE, VERSION);
	else
		udm_snprintf(passwd_tmp, 32, "%s", passwd);
	
	len = strlen(user_tmp)+strlen("USER ");
	buf = (char *) UdmXmalloc(len+1);
	udm_snprintf(buf, len+1, "USER %s", user_tmp);
	code = ftp_send_cmd(connp, buf);
	UDM_FREE(buf);
	if (code == -1)
		return -1;
	else if (code == 2) /* Don't need password */
		return 0;
		
	len = strlen(passwd_tmp)+strlen("PASS ");
	buf = (char *) UdmXmalloc(len+1);
	udm_snprintf(buf, len+1, "PASS %s", passwd_tmp);
	code = ftp_send_cmd( connp, buf);
	UDM_FREE(buf);
	if (code > 3)
		return -1;
	return 0;
}

int ftp_parse_list(UDM_CONN *connp, char *path){
	char *line, *buf_in, *ch, *buf_out, *tok, *fname;
        int len, len_h, len_f,len_p, cur_len, buf_len, i;
	
	if (!connp->buf || !connp->buf_len)
		return 0;
        buf_in = connp->buf;
        len_h = strlen(connp->hostname)+22;
        len_p = strlen(path);
        cur_len = 0;
        buf_len = NET_BUF_SIZE;
        buf_out = UdmXmalloc(buf_len);
	line=UdmGetToken(buf_in,"\r\n",&tok);
        do{
    		if (!(fname = strtok(line, " ")))
			continue;
		/* drwxrwxrwx x user group size month date time file_name */
		for(i=0; i<7; i++)
            		if (!(fname = strtok(NULL, " ")))
				break;
		if (!(fname = strtok(NULL, "")))
			continue;
		len = 0 ;
		len_f = len_h + len_p + strlen(fname);
	        if ((cur_len+len_f) >= buf_len){
			buf_len += NET_BUF_SIZE;
			buf_out = UdmXrealloc(buf_out, buf_len);
		}
					
		switch (line[0]){
			case 'd':
				if (!fname || !strcmp(fname, ".") || !strcmp(fname, ".."))
				        break;
				len = len_f;
    				udm_snprintf(buf_out+cur_len, len+1, "<a href=\"ftp://%s%s%s/\"></a>",
					    connp->hostname, path, fname);
				break;
    	    	        case 'l':
				ch = strstr (fname, " -> ");
				if (!ch)
				    break;
				ch +=4;
				len = len_h+len_p+strlen(ch);
    				udm_snprintf(buf_out+cur_len, len+1, "<a  href=\"ftp://%s%s%s\"></a>", 
					    connp->hostname, path, ch);
				break;
    	    	        case '-':
				len =  len_f; 
		    	        udm_snprintf(buf_out+cur_len, len+1, "<a  href=\"ftp://%s%s%s\"></a>", 
					    connp->hostname, path, fname);

				break;
		}
		cur_len += len;
		
	}while( (line=UdmGetToken(NULL, "\r\n", &tok)));

	if (cur_len+1 > connp->buf_len_total){
		connp->buf_len_total = cur_len;  
		connp->buf = UdmXrealloc(connp->buf, connp->buf_len_total+1);
	}
	bzero(connp->buf, connp->buf_len_total+1);
	memcpy(connp->buf, buf_out, cur_len);
	UDM_FREE(buf_out);
	return 0;
}

int ftp_list(UDM_CONN *c, UDM_CONN *d, char *path, char *filename, int max_doc_size){
	char *cmd;
	int len;
	
	if (!filename){
		cmd = UdmXmalloc(5);
		sprintf(cmd, "LIST");
	}else{
    		len = strlen(filename)+5;
		cmd = UdmXmalloc(len+1);
    		udm_snprintf(cmd, len+1, "LIST %s", filename);
	}

	if (ftp_send_data_cmd(c, d, cmd,max_doc_size)== -1){
		UDM_FREE(cmd);
		UdmLog(c->indexer, UDM_LOG_DEBUG, "(ftp_list-err)->%s", d->buf);
		return -1;
	}
	UDM_FREE(cmd);
	ftp_parse_list(d, path);
	return 0;
}

int ftp_get(UDM_CONN *c, UDM_CONN *d, char *path, int max_doc_size){
	char *cmd;
	int len;
	
        if (!path)
		return -1;

        len = strlen(path)+5;
	cmd = UdmXmalloc(len+1);
        udm_snprintf(cmd, len+1, "RETR %s", path);

	if (ftp_send_data_cmd(c, d, cmd,max_doc_size) == -1){
		UDM_FREE(cmd);
		return -1;
	}
	UDM_FREE(cmd);
	return 0;
}

int ftp_mdtm(UDM_CONN *c, char *path){
	char *cmd;
	int len, code;
	
	if (!path)
		return -1;
		
    	len = strlen(path)+5;
	cmd = UdmXmalloc(len+1);
	udm_snprintf(cmd, len+1, "MDTM %s", path);

	code = ftp_send_cmd(c, cmd);
	UDM_FREE(cmd);
	if (code == -1){
                return -1;
	}else if ( code >3 ){
		c->err = code;
		return -1;
	}
	return (UdmFTPDate2Time_t(c->buf));
}

int ftp_set_binary(UDM_CONN *c){
	char *cmd;
	int code;
	
	cmd = UdmXmalloc(7);
	sprintf(cmd, "TYPE I");

	code = ftp_send_cmd(c, cmd);
	UDM_FREE(cmd);
	if (code == -1){
                return -1;
	}else if ( code >3 ){
		c->err = code;
		return -1;
	}
	return 0;
}

int ftp_cwd(UDM_CONN *c, char *path){
	char *cmd;
	int len, code;
	
	if (!path)
		return -1;
		
	len = strlen(path)+4;
	cmd = UdmXmalloc(len+1);
	udm_snprintf(cmd, len+1, "CWD %s", path);

	code = ftp_send_cmd(c, cmd);
	UDM_FREE(cmd);
	if (code == -1){
                return -1;
	}else if ( code >3 ){
		c->err = code;
		return -1;
	}
	return 0;
}

int ftp_close(UDM_CONN *connp){
	int code;

	if (connp->connected == NET_CONNECTED){
    		code = ftp_send_cmd(connp, "QUIT");
	}
	connp->connected = NET_NOTCONNECTED;
	socket_close(connp);
	if (connp->connp)
		socket_close(connp->connp);
	return 0;
}

int ftp_abort(UDM_CONN *connp){
	int code;

	socket_buf_clear(connp->connp);
	if (send(connp->conn_fd, "\xFF\xF4\xFF", 3, MSG_OOB)==-1)
		return -1;
	
	if (socket_write(connp, "\xF2"))
		return -1;
	
	code = ftp_send_cmd(connp, "ABOR");
	socket_buf_clear(connp->connp);

	if (code !=4)
		return -1;
	return 0;
}
#endif
